from django.apps import AppConfig
from django.utils.translation import pgettext_lazy

from .discovery import autodiscover


__all__ = ('TwoFactorConfig',)


class TwoFactorConfig(AppConfig):
    name = 'wcd_2factor'
    verbose_name = pgettext_lazy('wcd_2factor', 'Two factor')

    def ready(self):
        autodiscover()



import random
from typing import Optional, Tuple
from uuid import UUID
import logging

from ..conf import settings
from ..models import ConfirmationState
from ..utils import cached_import_string


logger = logging.getLogger(__name__)


def make_confirmation_code() -> str:
    items = ['1', '2', '3', '4', '5', '6', '7', '8', '9'] * 3
    random.shuffle(items)

    return ''.join(items[:6])


def make_confirmation(
    is_confirmed: bool = False,
    meta: Optional[dict] = None
) -> ConfirmationState:
    code = cached_import_string(settings.CONFIRM_CODE_GENERATOR)()

    logger.debug(f'2Factor confirmation code: {code}')

    return ConfirmationState.objects.create(
        code=code, is_confirmed=is_confirmed, meta=meta or {}
    )


def confirm(id: UUID, code: str) -> Tuple[Optional[ConfirmationState], bool]:
    state: Optional[ConfirmationState] = ConfirmationState.objects.filter(id=id, code=code).first()

    if state is None:
        return None, False

    state.confirm()

    return state, state.is_confirmed


def check(id: UUID) -> Tuple[Optional[ConfirmationState], bool]:
    state = ConfirmationState.objects.filter(id=id).first()

    return state, state.is_confirmed if state is not None else False


def use(state: ConfirmationState) -> bool:
    if not state.is_confirmed:
        return False

    state.delete()

    return True



from django.contrib import admin

from .models import ConfirmationState


@admin.register(ConfirmationState)
class ConfirmationStateAdmin(admin.ModelAdmin):
    list_display = 'id', 'code', 'is_confirmed', 'created_at', 'updated_at',
    date_hierarchy = 'created_at'
    list_filter = 'is_confirmed', 'created_at', 'updated_at'
    search_fields = 'code', 'meta',



from django.utils.translation import pgettext
from rest_framework.serializers import CharField
from rest_framework.exceptions import ValidationError
from rest_framework_simplejwt.serializers import TokenObtainPairSerializer

from wcd_2factor.services import confirmer


class TwoFactorTokenObtainPairSerializer(TokenObtainPairSerializer):
    two_factor_id = CharField(required=True)

    def validate(self, attrs):
        data = super().validate(attrs)
        two_factor_id = attrs['two_factor_id']
        state, confirmed = confirmer.check(two_factor_id)

        if not confirmed:
            raise ValidationError({
                'two_factor_id': pgettext('wcd_2factor', 'Wrong confirmation.'),
            })

        confirmer.use(state)
        self.state = state

        return data



from typing import Type
from .config import SenderConfig
from .backend import SenderBackend
from ..utils import cached_import_string


__all__ = 'resolve_backend_from_config',


def resolve_backend_from_config(name: str, config: SenderConfig, kwargs: dict = {}):
    BackendClass: Type[SenderBackend] = cached_import_string(config['backend'])

    return BackendClass(
        name,
        options=config.get('options'),
        verbose_name=config.get('verbose_name'),
        **kwargs
    )



from typing import Optional
from ..conf import settings
from ..sender import resolve_backend_from_config, SenderBackend
from ..models import ConfirmationState


def get_sender_backend(
    sender_name: Optional[str] = None, **kwargs
) -> SenderBackend:
    name: str = sender_name or settings.DEFAULT_SENDER
    config = settings.SENDERS[name]

    return resolve_backend_from_config(name, config, kwargs)


def send(way: Optional[str], token: str, state: ConfirmationState, context: dict = {}):
    sender = get_sender_backend(way)

    return sender(token=token, state=state, context=context)



import graphene
from django.utils.translation import get_language
from django.utils.translation import pgettext_lazy
from wcd_graphene.exceptions import StatefulError
from postie.shortcuts import send_mail

from wcd_2factor.cases import confirmer


__all__ = 'RequestConfirmation', 'Confirm',

# TODO: Everything wrong here. Rewrite and do not document for now.
class RequestConfirmation(graphene.Mutation):
    id = graphene.String()
    # FIXME: Remove code from here in future, after test perhaps.
    code = graphene.String()

    class Arguments:
        token = graphene.String(required=True)
        way = graphene.String(required=True)

    def mutate(root, info, token: str, way: str):
        state = confirmer.make_confirmation()

        if way == 'email':
            send_mail(
                event='confirm_email',
                recipients=[token],
                context={'code': state.code},
                language=get_language()
            )
        if way == 'phone':
            # FIXME: Blocked for a while
            # from apps.staff.tasks import send_phone_confirmation
            # send_phone_confirmation.delay(phone=token, code=state.code)
            pass

        return RequestConfirmation(id=str(state.id), code=state.code)


class Confirm(graphene.Mutation):
    id = graphene.String()

    class Arguments:
        id = graphene.String(required=True)
        code = graphene.String(required=True)

    def mutate(root, info, id: str, code: str):
        confirmed, state = confirmer.confirm(id, code)

        if not confirmed:
            raise StatefulError(
                message=pgettext_lazy('staff', 'Wrong code'),
                extend_path=['code']
            )

        return Confirm(id=id)



# Generated by Django 3.2.8 on 2022-01-12 12:05

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ConfirmationState',
            fields=[
                ('id', models.UUIDField(primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.TextField(verbose_name='Confirmation code')),
                ('is_confirmed', models.BooleanField(default=False, max_length=512, verbose_name='Confirmed')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated at')),
            ],
            options={
                'verbose_name': 'Confirmation state',
                'verbose_name_plural': 'Confirmation states',
            },
        ),
    ]



from django.dispatch import Signal


__all__ = 'code_sent', 'code_confirmed',


# Arguments: "user_id", "code"
code_sent: Signal = Signal()

# Arguments: "user_id", "code"
code_confirmed: Signal = Signal()



from django.urls import path
from rest_framework.generics import GenericAPIView
from rest_framework import status
from rest_framework.response import Response
from django.db import transaction

from wcd_2factor.conf import settings
from .serializers import RequestConfirmationSerializer, ConfirmSerializer


__all__ = (
    'RequestConfirmationView', 'ConfirmView',
    'request_confirmation_view', 'confirm_view', 'make_urlpatterns',
)


class SerializerCommitterMixin:
    def collect_response(self, request, state, **kwargs):
        response = {'id': state.id}

        if settings.DEBUG_CODE_RESPONSE:
            response['code'] = state.code

        return response

    @transaction.atomic()
    def post(self, request, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        state = serializer.commit()

        return Response(
            self.collect_response(request, serializer=serializer, state=state),
            status=status.HTTP_200_OK,
        )


class RequestConfirmationView(SerializerCommitterMixin, GenericAPIView):
    serializer_class = RequestConfirmationSerializer


class ConfirmView(SerializerCommitterMixin, GenericAPIView):
    serializer_class = ConfirmSerializer


request_confirmation_view = RequestConfirmationView.as_view()
confirm_view = ConfirmView.as_view()


def make_urlpatterns(
    request_confirmation_view=request_confirmation_view,
    confirm_view=confirm_view,
):
    return [
        path('request-confirmation/', request_confirmation_view, name='request-confirmation'),
        path('confirm/', confirm_view, name='confirm'),
    ]



from uuid import uuid4
from django.db import models
from django.utils.translation import pgettext_lazy, pgettext
from django.contrib.postgres.fields import JSONField

from .query import ConfirmationStateQuerySet


__all__ = 'ConfirmationState',


class ConfirmationState(models.Model):
    objects: models.Manager = (
        ConfirmationStateQuerySet.as_manager()
    )

    class Meta:
        verbose_name = pgettext_lazy('wcd_2factor', 'Confirmation state')
        verbose_name_plural = pgettext_lazy(
            'wcd_2factor', 'Confirmation states'
        )

    id = models.UUIDField(
        primary_key=True, verbose_name=pgettext_lazy('wcd_2factor', 'ID'),
        default=uuid4
    )
    code = models.TextField(
        verbose_name=pgettext_lazy('wcd_2factor', 'Confirmation code'),
        blank=False, null=False,
    )

    is_confirmed = models.BooleanField(
        verbose_name=pgettext_lazy('wcd_2factor', 'Confirmed'), max_length=512,
        null=False, blank=False, default=False
    )

    meta = JSONField(
        verbose_name=pgettext_lazy('wcd_2factor', 'Metadata'),
        null=False, blank=True, default=dict
    )

    created_at = models.DateTimeField(
        verbose_name=pgettext_lazy('wcd_2factor', 'Created at'),
        auto_now_add=True,
    )
    updated_at = models.DateTimeField(
        verbose_name=pgettext_lazy('wcd_2factor', 'Updated at'), auto_now=True,
    )

    def __str__(self):
        return (
            pgettext(
                'wcd_2factor',
                '#{code} confirmation: {confirmed}.',
            )
            .format(
                code=self.code,
                confirmed='+' if self.is_confirmed else '-',
            )
        )

    def confirm(self):
        self.is_confirmed = True
        self.save(update_fields=('is_confirmed',))



from dataclasses import dataclass, field
from typing import Dict, TYPE_CHECKING
from django.conf import settings as django_settings

from px_settings.contrib.django import settings as s

if TYPE_CHECKING:
    from .sender import SenderConfig


__all__ = 'Settings', 'settings',


@s('WCD_2FACTOR')
@dataclass
class Settings:
    """
    Example:

    ```python
    WCD_2FACTOR = {
        'SENDERS': {
            'default': {
                'verbose_name': 'Phone sender',
                'backend': 'some.method.path.to.Backend',
                'options': {
                    'SECRET: 'SECRET',
                },
            },
        },
        'DEFAULT_SENDER': 'default',
        'CONFIRM_CODE_GENERATOR': 'wcd_2factor.services.confirmer.make_confirmation_code',
        'DEBUG_CODE_RESPONSE': False,
    }
    ```
    """
    _DEBUG_CODE_RESPONSE: bool = False
    DEBUG_CODE_RESPONSE: bool = _DEBUG_CODE_RESPONSE

    SENDERS: Dict[str, 'SenderConfig'] = field(default_factory=dict)
    DEFAULT_SENDER: str = 'default'
    CACHE: str = 'default'
    CONFIRM_CODE_GENERATOR: str = 'wcd_2factor.services.confirmer.make_confirmation_code'

    def get_sender_choices(self):
        return [
            (key, config.get('verbose_name', key))
            for key, config in self.SENDERS.items()
        ]

    @property
    def DEBUG_CODE_RESPONSE(self) -> bool:
        return self._DEBUG_CODE_RESPONSE and django_settings.DEBUG

    @DEBUG_CODE_RESPONSE.setter
    def DEBUG_CODE_RESPONSE(self, value: bool) -> None:
        self._DEBUG_CODE_RESPONSE = value


settings = Settings()



# Generated by Django 2.2 on 2022-04-06 14:36

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('wcd_2factor', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='confirmationstate',
            name='meta',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict, verbose_name='Metadata'),
        ),
        migrations.AlterField(
            model_name='confirmationstate',
            name='id',
            field=models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]



from typing import Any, Dict, Optional, TypedDict

from .backend import SenderOptions


__all__ = 'SenderConfig',


class SenderConfig(TypedDict):
    backend: str
    options: Optional[SenderOptions]
    verbose_name: Optional[str]



class TwoFactorError(Exception):
    pass


class WrongBackend(TwoFactorError):
    backend_name: str

    def __init__(self, backend_name):
        self.backend_name = backend_name

    def __str__(self):
        return f'Incorrect 2factor backend: "{self.backend_name}".'


class MissingBackend(WrongBackend):
    def __str__(self):
        return f'No such backend exists: "{self.backend_name}".'



from rest_framework.exceptions import ValidationError
from django.utils.translation import pgettext
from rest_framework.serializers import Serializer, CharField, ChoiceField
from django.utils.functional import SimpleLazyObject

from wcd_2factor.conf import settings
from wcd_2factor.services import confirmer, sender


class RequestConfirmationSerializer(Serializer):
    token = CharField(required=True)
    way = ChoiceField(
        choices=SimpleLazyObject(lambda: settings.get_sender_choices()),
        required=False,
    )

    def commit(self):
        state = confirmer.make_confirmation(meta={
            'token': self.validated_data['token'],
        })

        sender.send(
            self.validated_data.get('way'),
            self.validated_data['token'],
            state,
            context={'request': self.context['request']},
        )

        return state


class ConfirmSerializer(Serializer):
    id = CharField(required=True)
    code = CharField(required=True)

    def validate_id(self, value):
        state, confirmed = confirmer.check(value)

        if state is None:
            raise ValidationError(
                pgettext('wcd_2factor', 'Wrong confirmation id.')
            )

        if confirmed:
            raise ValidationError(
                pgettext('wcd_2factor', 'Confirmation is already confirmed.')
            )

        return state

    def validate(self, data):
        data = super().validate(data)

        if data['id'].code != data['code']:
            raise ValidationError({
                'code': pgettext('wcd_2factor', 'Wrong confirmation code.'),
            })

        return data

    def commit(self):
        state = self.validated_data['id']
        state.confirm()

        return state



from typing import Any, Callable, Dict, Optional
from django.utils.text import camel_case_to_spaces

from ..models import ConfirmationState


__all__ = 'SenderBackend', 'FunctionalSenderBackend',


SenderOptions = Dict[str, Any]


class SenderBackend:
    name: str
    options: SenderOptions
    verbose_name: str

    def __init__(
        self,
        name: str,
        options: Optional[SenderOptions] = None,
        verbose_name: Optional[str] = None,
        **kwargs
    ) -> None:
        self.name = name
        self.options = options or {}
        self.verbose_name = (
            verbose_name or camel_case_to_spaces(self.__class__.__name__)
        )
        self.kwargs = kwargs

    def send(
        self,
        token: str,
        state: ConfirmationState,
        context: dict = {}
    ): # pragma: no cover
        raise NotImplementedError(
            'You should implement `send` method in your sender backend'
        )

    def __call__(self, *args, **kwargs):
        return self.send(*args, **kwargs)


class FunctionalSenderBackend(SenderBackend):
    def __init__(self, func: Callable, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.functional_send = func

    def send(
        self,
        token: str,
        state: ConfirmationState,
        context: dict = {}
    ):
        return self.functional_send(
            token=token, state=state, context=context, name=self.name,
            options=self.options, **self.kwargs
        )

    @classmethod
    def from_callable(cls, func):
        def wrapper(*args, **kwargs):
            return cls(func, *args, **kwargs)

        return wrapper
